#define BASE	0x00008000

	.text

	.align	2

start:
	.word	0x464c457f			// e_ident
	.word	0x61010101
	.word	0x00000000
	.word	0x00000000
	.short	0x0002				// e_type
	.short	0x0028				// e_machine
	.word	0x00000001			// (e_version)
	.word	entry-start+BASE	// e_entry
	.word	0x0000002c			// e_phoff
	.word	0x00000000			// (e_shoff)
	.word	0x00000202			// (e_flags)
	.short	0x0034				// (e_ehsize)
	.short	0x0020				// e_phentsize
	.short	0x0001				// e_phnum			// p_type
	.short	0x0000				// (e_shentsize)
	.short	0x0000				// (e_shnum)		// p_offset
	.short	0x0000				// (e_shstrndx)
	.word	BASE									// p_vaddr
	.word	0x00008000								// (p_paddr)
	.word	end-start								// p_filesz
	.word	0x00000200								// (p_memsz)
	.word	0x00000007								// p_flags
	.word	0x00008000								// (p_align)

#define BAT_WIDTH        32
#define BAT_WIDTH_SHIFT	  5
#define BAT_Y           220

dev_mem_filename:
	.ascii "/dev/mem"

reglist:
	.word	0x290c0200		// store 512    in gpio 0x290c
	.word	0x28da02ab		// store 0x02ab in gpio 0x28da
	.word	0x29580000		// set palette starting at entry 0
	.word	0x295affff
	.word	0x295affff
	.word	0x295a0000
	.word	0x295a0000
	.word	0x295affff
	.word	0x295a0000

mmap_params:
	.word	0				// addr
	.word	0x1e000			// size
	.word	3				// prot = PROT_READ | PROT_WRITE
	.word	1				// flags = MAP_SHARED
mmap_params_fd:
	.word	4				// fd
mmap_params_offset:
	.word	0x03101000		// offset

ball_bat_color_x4:
	.word	0x01010101

	.align	2
	.global	entry
entry:
	// open /dev/mem
	adr		r0, dev_mem_filename
	mov		r1, #2				// O_RDWR
	swi		#0x900005			// sys_open

	// 'open' is documented to return the smallest fd not already opened
	// by the process, so this will always be 3 - no need to actually store
	// the result then!
	//str	r0, mmap_params_fd

	// mmap the vram
	adr		r0, mmap_params
	swi		#0x90005a			// old_mmap
	mov		r11, r0				// r11 = vram addr

	// mmap the gpio regs
	adr		r0, mmap_params
	mov		r2, #0xc0000000
	str		r2, mmap_params_offset
	swi		#0x90005a			// old_mmap
	mov		r10, r0				// r10 = gpio addr

	add		r4, r10, #0x1180	// useful reference point


	// set gpio regs
	adr		r0, reglist
1:
	ldr		r1, [r0], #4
	movs	r7, r1, lsr #16
	strneh	r1, [r10, r7]
	bne		1b


	// initialise ball and bat data
	mov		r8, #4						// ball x velocity
	mov		r6, #160					// ball x coord
	mov		r12, #160-BAT_WIDTH/2		// bat x coord

init:

	// A little delay loop... but on startup, r7 is zero so it exits quickly.
	// Later on when we get back here, r7=236, so this loop waits about 1 second.
1:
	bl		vsync
	subs	r7, r7, #4
	bge		1b

	// reset some of the data
	mov		r7, #BAT_Y - 32				// ball y coord
	mov		r9, #-4						// ball y velocity
	mov		r5, #0						// ceiling height


main_loop:

	// Simple background
	add		r0, r11, #0x1e000		// end of screen memory
	add		r2, r11, r5, asl #9		// end of ceiling
	mov		r1, #2

bg_loop_y:
	sub		r0, r0, #512			// move up a line
	mov		r3, #320				// line length

bg_loop_x:
	subs	r3, r3, #4				// move left
	str		r1, [r0, r3]			// write a word
	bne		bg_loop_x				// loop until written zero offset

	cmp		r0, r2					// reached ceiling?
	ldreq	r1, ball_bat_color_x4

	mov		r1, r1, ror #16
	cmp		r0, r11					// start of screen memory
	bne		bg_loop_y


	// draw ball and bat
	//ldr		r1, ball_bat_color_x4	// already loaded
	
	// ball
	add		r0, r6, r7, asl #9    // ball screen offset
	add		r0, r0, r11
	bic		r2, r1, #0x00000001
	bic		r2, r2, #0x01000000
	str		r2, [r0]
	str		r1, [r0, #512]!
	str		r1, [r0, #512]!
	str		r2, [r0, #512]

	// bat
	add		r0, r12, #BAT_Y*512   // bat screen offset
	add		r0, r0, r11
	mov		r2, #BAT_WIDTH/4
1:
	str		r1, [r0, #512]
	str		r1, [r0], #4
	subs	r2, r2, #1
	bne		1b


	// game over?
	cmp		r7, #232
	bge		init


	// sample stick inputs and move bat
	ldrh	r1, [r4, #24]		// read register M
	mvn		r1, r1				// flip bits, they're backwards
	tst		r1, #0x0E
	cmpne	r12, #0
	subne	r12, r12, #4
	tst		r1, #0xE0
	cmpne	r12, #320-BAT_WIDTH
	addne	r12, r12, #4

	// move ball horiz
	adds	r6, r6, r8
	cmpne	r6, #316
	rsbeq	r8, r8, #0
	
	// move ball vert
	add		r7, r7, r9			// move, hit roof?
	cmp		r7, r5
	beq		1f
	cmp		r7, #BAT_Y-4		// hit bat?
	subeq	r0, r6, r12			// get ball x rel to bat
	moveqs	r1, r0, lsr #BAT_WIDTH_SHIFT	// divide by bat width, compare against zero
	addeq	r5, r5, #8
	bne		2f
	tst		r0, #12				// sometimes do weird bounces
	bne		1f

	tst		r9,	#10				// force to 4 if already weird
	movne	r9, #4
	bne		1f

	tst		r6, #32				// or do a random change
	movne	r9, #2
	moveq	r9, #8

1:
	rsb		r9, r9, #0
2:

#if 0
	bl		vsync
	b		main_loop
#else
	adr		lr, main_loop
#endif


#if 0

vsync:
1:
	ldr		r1, [r4, #2]
	tst		r1, #0x10
	bne		1b
1:
	ldr		r1, [r4, #2]
	tst		r1, #0x10
	beq		1b
	mov		pc, lr

#else

vsync:
	adr		r2, vsync_loop
vsync_loop:
	ldr		r1, [r4, #2]
	tst		r1, #0x10
	movne	pc, r2
	mov		r2, lr
	b		vsync_loop

#endif

end:
